package javax.swing.text;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.awt.Font;
import java.util.Set;
import java.text.AttributedCharacterIterator;
import java.awt.Container;
import java.awt.font.FontRenderContext;
import java.awt.font.TextAttribute;
import javax.swing.JComponent;
import java.text.BreakIterator;
import sun.swing.SwingUtilities2;
import java.awt.font.TextLayout;
import sun.font.BidiUtils;
import java.awt.Rectangle;
import javax.swing.event.DocumentEvent;
import java.awt.font.LineBreakMeasurer;

class TextLayoutStrategy extends FlowView.FlowStrategy
{
    private LineBreakMeasurer measurer;
    private TextLayoutStrategy.AttributedSegment text;
    
    public TextLayoutStrategy() {
        this.text = new TextLayoutStrategy.AttributedSegment();
    }
    
    @Override
    public void insertUpdate(FlowView flowView, DocumentEvent documentEvent, Rectangle rectangle) {
        this.sync(flowView);
        super.insertUpdate(flowView, documentEvent, rectangle);
    }
    
    @Override
    public void removeUpdate(FlowView flowView, DocumentEvent documentEvent, Rectangle rectangle) {
        this.sync(flowView);
        super.removeUpdate(flowView, documentEvent, rectangle);
    }
    
    @Override
    public void changedUpdate(FlowView flowView, DocumentEvent documentEvent, Rectangle rectangle) {
        this.sync(flowView);
        super.changedUpdate(flowView, documentEvent, rectangle);
    }
    
    @Override
    public void layout(FlowView flowView) {
        super.layout(flowView);
    }
    
    @Override
    protected int layoutRow(FlowView flowView, int n, int n2) {
        int layoutRow = super.layoutRow(flowView, n, n2);
        View view = flowView.getView(n);
        Object property = flowView.getDocument().getProperty("i18n");
        if (property != null && property.equals(Boolean.TRUE)) {
            int viewCount = view.getViewCount();
            if (viewCount > 1) {
                Element bidiRootElement = ((AbstractDocument)flowView.getDocument()).getBidiRootElement();
                byte[] levels = new byte[viewCount];
                View[] objects = new View[viewCount];
                for (int i = 0; i < viewCount; ++i) {
                    View view2 = view.getView(i);
                    levels[i] = (byte)StyleConstants.getBidiLevel(bidiRootElement.getElement(bidiRootElement.getElementIndex(view2.getStartOffset())).getAttributes());
                    objects[i] = view2;
                }
                BidiUtils.reorderVisually(levels, objects);
                view.replace(0, viewCount, objects);
            }
        }
        return layoutRow;
    }
    
    @Override
    protected void adjustRow(FlowView flowView, int n, int n2, int n3) {
    }
    
    @Override
    protected View createView(FlowView flowView, int n, int n2, int n3) {
        View logicalView = this.getLogicalView(flowView);
        flowView.getView(n3);
        int requireNextWord = this.viewBuffer.size() != 0;
        View view = logicalView.getView(logicalView.getViewIndex(n, Position.Bias.Forward));
        int limitingOffset = this.getLimitingOffset(view, n, n2, (boolean)requireNextWord);
        if (limitingOffset == n) {
            return null;
        }
        if (n == view.getStartOffset() && limitingOffset == view.getEndOffset()) {
        }
        else {
            view = view.createFragment(n, limitingOffset);
        }
        if (view instanceof GlyphView && this.measurer != null) {
            boolean n4 = false;
            int startOffset = view.getStartOffset();
            int endOffset = view.getEndOffset();
            if (endOffset - startOffset == 1 && ((GlyphView)view).getText(startOffset, endOffset).first() == '\t') {
                n4 = true;
            }
            TextLayout textLayout = n4 ? null : this.measurer.nextLayout((float)n2, this.text.toIteratorIndex(limitingOffset), (boolean)requireNextWord);
            if (textLayout != null) {
                ((GlyphView)view).setGlyphPainter(new GlyphPainter2(textLayout));
            }
        }
        return view;
    }
    
    int getLimitingOffset(View view, int n, int n2, boolean requireNextWord) {
        int b = view.getEndOffset();
        Document document = view.getDocument();
        if ((document instanceof AbstractDocument)) {
            Element bidiRootElement = ((AbstractDocument)document).getBidiRootElement();
            if (bidiRootElement.getElementCount() > 1) {
                b = Math.min(bidiRootElement.getElement(bidiRootElement.getElementIndex(n)).getEndOffset(), b);
            }
        }
        if ((view instanceof GlyphView)) {
            Segment text = ((GlyphView)view).getText(n, b);
            if (text.first() == '\t') {
                b = n + 1;
            }
            else {
                int i = text.next();
                while (i != '\uffff') {
                    if (i == '\t') {
                        b = n + text.getIndex() - text.getBeginIndex();
                        break;
                    }
                    i = text.next();
                }
            }
        }
        int offsetLimit = this.text.toIteratorIndex(b);
        if (this.measurer != null) {
            int iteratorIndex = this.text.toIteratorIndex(n);
            if (this.measurer.getPosition() != iteratorIndex) {
                this.measurer.setPosition(iteratorIndex);
            }
            offsetLimit = this.measurer.nextOffset((float)n2, offsetLimit, requireNextWord);
        }
        return this.text.toModelPosition(offsetLimit);
    }
    
    void sync(FlowView flowView) {
        View logicalView = this.getLogicalView(flowView);
        this.text.setView(logicalView);
        FontRenderContext fontRenderContext = SwingUtilities2.getFontRenderContext(flowView.getContainer());
        Container container = flowView.getContainer();
        BreakIterator breakIter;
        if (container != null) {
            breakIter = BreakIterator.getLineInstance(container.getLocale());
        }
        else {
            breakIter = BreakIterator.getLineInstance();
        }
        Object clientProperty = null;
        if ((container instanceof JComponent)) {
            clientProperty = ((JComponent)container).getClientProperty(TextAttribute.NUMERIC_SHAPING);
        }
        TextLayoutStrategy.AttributedSegment.access$000(this.text, clientProperty);
        this.measurer = new LineBreakMeasurer(this.text, breakIter, fontRenderContext);
        for (int viewCount = logicalView.getViewCount(), i = 0; i < viewCount; ++i) {
            View view = logicalView.getView(i);
            if ((view instanceof GlyphView)) {
                int startOffset = view.getStartOffset();
                int endOffset = view.getEndOffset();
                this.measurer.setPosition(this.text.toIteratorIndex(startOffset));
                ((GlyphView)view).setGlyphPainter(new GlyphPainter2(this.measurer.nextLayout(Float.MAX_VALUE, this.text.toIteratorIndex(endOffset), false)));
            }
        }
        this.measurer.setPosition(this.text.getBeginIndex());
    }
    
    static class AttributedSegment extends Segment implements AttributedCharacterIterator
    {
        View v;
        static Set<AttributedCharacterIterator.Attribute> keys;
        private Object shaper;
        
        AttributedSegment() {
            this.shaper = null;
        }
        
        View getView() {
            return this.v;
        }
        
        void setView(View v) {
            this.v = v;
            Document document = v.getDocument();
            int startOffset = v.getStartOffset();
            int endOffset = v.getEndOffset();
            try {
                document.getText(startOffset, endOffset - startOffset, this);
            }
            catch (BadLocationException ex) {
                throw new IllegalArgumentException("Invalid view");
            }
            this.first();
        }
        
        int getFontBoundary(int n, int n2) {
            View view = this.v.getView(n);
            Font font;
            for (font = this.getFont(n), n += n2; n >= 0 && n < this.v.getViewCount() && this.getFont(n) == font; n += n2) {
                view = this.v.getView(n);
            }
            return (n2 < 0) ? view.getStartOffset() : view.getEndOffset();
        }
        
        Font getFont(int n) {
            View view = this.v.getView(n);
            if ((view instanceof GlyphView)) {
                return ((GlyphView)view).getFont();
            }
            return null;
        }
        
        int toModelPosition(int n) {
            return this.v.getStartOffset() + (n - this.getBeginIndex());
        }
        
        int toIteratorIndex(int n) {
            return n - this.v.getStartOffset() + this.getBeginIndex();
        }
        
        private void setShaper(Object shaper) {
            this.shaper = shaper;
        }
        
        @Override
        public int getRunStart() {
            return this.toIteratorIndex(this.v.getView(this.v.getViewIndex(this.toModelPosition(this.getIndex()), Position.Bias.Forward)).getStartOffset());
        }
        
        @Override
        public int getRunStart(AttributedCharacterIterator.Attribute attribute) {
            if ((attribute instanceof TextAttribute)) {
                int viewIndex = this.v.getViewIndex(this.toModelPosition(this.getIndex()), Position.Bias.Forward);
                if (attribute == TextAttribute.FONT) {
                    return this.toIteratorIndex(this.getFontBoundary(viewIndex, -1));
                }
            }
            return this.getBeginIndex();
        }
        
        @Override
        public int getRunStart(Set<? extends AttributedCharacterIterator.Attribute> set) {
            int n = this.getBeginIndex();
            Object[] array = set.toArray();
            for (int i = 0; i < array.length; ++i) {
                n = Math.max(this.getRunStart((TextAttribute)array[i]), n);
            }
            return Math.min(this.getIndex(), n);
        }
        
        @Override
        public int getRunLimit() {
            return this.toIteratorIndex(this.v.getView(this.v.getViewIndex(this.toModelPosition(this.getIndex()), Position.Bias.Forward)).getEndOffset());
        }
        
        @Override
        public int getRunLimit(AttributedCharacterIterator.Attribute attribute) {
            if ((attribute instanceof TextAttribute)) {
                int viewIndex = this.v.getViewIndex(this.toModelPosition(this.getIndex()), Position.Bias.Forward);
                if (attribute == TextAttribute.FONT) {
                    return this.toIteratorIndex(this.getFontBoundary(viewIndex, 1));
                }
            }
            return this.getEndIndex();
        }
        
        @Override
        public int getRunLimit(Set<? extends AttributedCharacterIterator.Attribute> set) {
            int n = this.getEndIndex();
            Object[] array = set.toArray();
            for (int i = 0; i < array.length; ++i) {
                n = Math.min(this.getRunLimit((TextAttribute)array[i]), n);
            }
            return Math.max(this.getIndex(), n);
        }
        
        @Override
        public Map<AttributedCharacterIterator.Attribute, Object> getAttributes() {
            Object[] array = TextLayoutStrategy.AttributedSegment.keys.toArray();
            Hashtable<TextAttribute, Object> hashtable = new Hashtable();
            for (int i = 0; i < array.length; ++i) {
                TextAttribute key = (TextAttribute)array[i];
                Object attribute = this.getAttribute(key);
                if (attribute != null) {
                    hashtable.put(key, attribute);
                }
            }
            return hashtable;
        }
        
        @Override
        public Object getAttribute(AttributedCharacterIterator.Attribute attribute) {
            int viewIndex = this.v.getViewIndex(this.toModelPosition(this.getIndex()), Position.Bias.Forward);
            if (attribute == TextAttribute.FONT) {
                return this.getFont(viewIndex);
            }
            if (attribute == TextAttribute.RUN_DIRECTION) {
                return this.v.getDocument().getProperty(TextAttribute.RUN_DIRECTION);
            }
            if (attribute == TextAttribute.NUMERIC_SHAPING) {
                return this.shaper;
            }
            return null;
        }
        
        @Override
        public Set<AttributedCharacterIterator.Attribute> getAllAttributeKeys() {
            return TextLayoutStrategy.AttributedSegment.keys;
        }
        
        static void access$000(TextLayoutStrategy.AttributedSegment attributedSegment, Object shaper) {
            attributedSegment.setShaper(shaper);
        }
        
        static {
            (TextLayoutStrategy.AttributedSegment.keys = new HashSet()).add(TextAttribute.FONT);
            TextLayoutStrategy.AttributedSegment.keys.add(TextAttribute.RUN_DIRECTION);
            TextLayoutStrategy.AttributedSegment.keys.add(TextAttribute.NUMERIC_SHAPING);
        }
    }
}
